import numpy as np
from sklearn.neighbors import NearestNeighbors
from scipy.sparse import issparse
from scipy.sparse import csr_matrix
from sklearn.utils import check_array
from sklearn.utils import check_random_state
from sklearn.utils import shuffle
from sklearn.preprocessing import LabelBinarizer

class MultiSMOTE:
    """Implementation of Multi-label SMOTE with vectorized operations for efficient processing.

    Parameters:
    -----------
    k_neighbors : int, default=5
        Number of nearest neighbors to use for generating synthetic samples.
    
    random_state : int, RandomState instance or None, default=None
        Controls the randomness of the sampling.
    
    n_jobs : int, default=None
        Number of CPU cores to use during the cross-validation.
    """
    
    def __init__(self, k_neighbors=5, random_state=None, n_jobs=None):
        self.k_neighbors = k_neighbors
        self.random_state = random_state
        self.n_jobs = n_jobs
        
    def _calculate_imbalance_ratios(self, y):
        """Calculate imbalance ratios for each label."""
        label_counts = np.sum(y, axis=0)
        total_samples = y.shape[0]
        irpl = (total_samples - label_counts) / label_counts
        mir = np.mean(irpl)
        return irpl, mir
        
    def _identify_tail_labels(self, irpl, mir):
        """Identify tail labels based on their imbalance ratio."""
        return np.where(irpl > mir)[0]
        
    def _generate_synthetic_samples(self, X, y, minority_indices, nn_indices):
        """Generate synthetic samples using vectorized operations."""
        random_state = check_random_state(self.random_state)
        
        # Get random neighbor indices for each minority sample
        rand_neighbors = random_state.randint(0, self.k_neighbors, len(minority_indices))
        neighbor_indices = nn_indices[minority_indices, rand_neighbors]
        
        # Vectorized generation of synthetic samples
        gaps = random_state.random_sample(len(minority_indices))
        X_diff = X[neighbor_indices] - X[minority_indices]
        X_syn = X[minority_indices] + gaps[:, None] * X_diff
        
        return X_syn
        
    def _assign_labels_ranking(self, y, minority_indices, nn_indices):
        """Assign labels using ranking approach."""
        # Get labels for minority samples and their neighbors
        y_minority = y[minority_indices]
        y_neighbors = y[nn_indices[minority_indices]]
        
        # Count label frequencies among neighbors
        label_counts = np.sum(y_neighbors, axis=1)
        
        # Create synthetic labels using ranking approach
        y_syn = np.zeros((len(minority_indices), y.shape[1]), dtype=int)
        y_syn[label_counts > self.k_neighbors // 2] = 1
        
        # Ensure at least one label is assigned
        y_syn[np.sum(y_syn, axis=1) == 0] = y_minority[np.sum(y_syn, axis=1) == 0]
        
        return y_syn
        
    def fit_resample(self, X, y):
        """Generate synthetic samples for the minority classes.
        
        Parameters:
        -----------
        X : array-like of shape (n_samples, n_features)
            The input samples.
            
        y : array-like of shape (n_samples, n_labels)
            The target labels.
            
        Returns:
        --------
        X_resampled : array-like of shape (n_samples_new, n_features)
            The array containing the input samples and the new synthetic samples.
            
        y_resampled : array-like of shape (n_samples_new, n_labels)
            The corresponding target labels.
        """
        X = check_array(X, accept_sparse=['csr'])
        y = check_array(y, accept_sparse=['csr'])
        
        if issparse(y):
            y = y.toarray()
        
        random_state = check_random_state(self.random_state)
        
        # Calculate imbalance ratios and identify tail labels
        irpl, mir = self._calculate_imbalance_ratios(y)
        tail_labels = self._identify_tail_labels(irpl, mir)
        
        # Find samples containing any tail label
        minority_indices = np.any(y[:, tail_labels], axis=1)
        minority_indices = np.where(minority_indices)[0]
        
        if len(minority_indices) < self.k_neighbors + 1:
            return X, y
        
        # Create nearest neighbors model
        nn = NearestNeighbors(n_neighbors=self.k_neighbors + 1, n_jobs=self.n_jobs)
        nn.fit(X)
        
        # Find nearest neighbors for all samples
        nn_indices = nn.kneighbors(X, return_distance=False)[:, 1:]
        
        # Generate synthetic samples
        X_syn = self._generate_synthetic_samples(X, y, minority_indices, nn_indices)
        
        # Assign labels using ranking approach
        y_syn = self._assign_labels_ranking(y, minority_indices, nn_indices)
        
        # Combine original and synthetic samples
        X_resampled = np.vstack([X, X_syn])
        y_resampled = np.vstack([y, y_syn])
        
        # Shuffle the samples
        X_resampled, y_resampled = shuffle(X_resampled, y_resampled, 
                                          random_state=random_state)
        
        return X_resampled, y_resampled
